"""Implementation details to interact with the serialization means."""

import datetime
import json
import math

import dolon.db_conn as db_conn
import dolon.exceptions as exceptions
import dolon.impl.constants as constants

# Aliases.
DbConnection = db_conn.DbConnection

_PREFETCH_SIZE = 100


async def process_message(db, payload):
    """Processes a tracing message storing it to the db.

    :param db: The database object to use.

    :param dict payload: A dict representing the message to store.

    Can be either a tracing run creation in the form of:

        msg = {
            "msg_type": "create_trace_run",
            "app_name": app_name,
            "uuid": identifier,
            "column_names": ["v1", 'v2']
        }

    or for the insertion of a tracing row:

        msg = {
            "msg_type": "row",
            "uuid": identifier,
            "row_data": [12.2, 123.1]
        }

    raises: InvalidMessage
    """
    if not isinstance(payload, dict):
        assert isinstance(payload, str)
        msg = json.loads(payload)
    else:
        msg = payload
    # logging.info(str(payload)) # Will cause missed messages.
    msg_type = msg.get('msg_type')
    try:
        if msg_type == "create_trace_run":
            identifier = msg.get('uuid')
            app_name = msg.get('app_name')
            column_names = msg.get('column_names')
            if not all([identifier, app_name, column_names]):
                raise exceptions.InvalidMessage(
                    f"Message not supported: {str(payload)}"
                )
            column_names = list(column_names)
            await _create_tracer(db, identifier, app_name, *column_names)
        elif msg_type == 'row':
            identifier = msg.get('uuid')
            row_data = msg.get('row_data')
            await _insert_row(db, identifier, *row_data)
        else:
            raise exceptions.InvalidMessage(
                f"Message not supported: {str(payload)}"
            )
    except Exception as ex:
        raise exceptions.InvalidMessage(
            f"Message not supported: {str(payload)}"
        ) from ex


async def get_trace_as_json(uuid):
    """Returns all the tracing rows for the passed in uuid as json.

    :param str uuid: The identifier for the trace run.

    :returns: All the tracing rows for the passed in uuid as json.
    :rtype: list[dict]
    """
    data = await get_trace(uuid)
    lines = data.split('\n')
    field_names = lines[0].split(',')
    columns = [list() for _ in range(len(field_names))]

    for line in lines[1:]:
        fields = line.split(',')
        for index, value in enumerate(fields):
            try:
                columns[index].append(float(value))
            except ValueError:
                columns[index].append(value)

    trace_as_json = {}
    for field_name, values in zip(field_names, columns):
        trace_as_json[field_name] = values

    return trace_as_json


async def get_trace_run_info(uuid):
    """Returns descriptive info for the passed in uuid.

    :param str uuid: The identifier for the trace run.

    :returns: Descriptive info for the passed in uuid.
    :rtype: dict
    """
    app_name = None
    counter = None
    from_time = None
    to_time = None

    async with DbConnection() as db:
        conn_pool = db.get_conn_pool()
        async with conn_pool.acquire() as conn:
            stmt = await conn.prepare(constants.SQL_SELECT_APP_NAME)
            async with conn.transaction():
                async for record in stmt.cursor(uuid,
                                                prefetch=_PREFETCH_SIZE):
                    app_name = record['app_name']
                    creation_time = _format_datetime(record['creation_time'])
            stmt = await conn.prepare(constants.SQL_SELECT_RUN_INFO)
            async with conn.transaction():
                async for record in stmt.cursor(uuid,
                                                prefetch=_PREFETCH_SIZE):
                    counter = record['counter']
                    from_time = record['from_time']
                    to_time = record['to_time']

    if to_time is None or from_time is None:
        # There are no rows for this run
        counter = 0
        started = creation_time
        duration = 'n/a'
    else:
        started = creation_time
        duration = _get_duration(to_time, from_time)
    return {
        'app_name': app_name,
        'counter': f'{counter:,}',
        'started': started,
        'duration': duration
    }


async def get_latest_trace(app_name):
    """Returns the latest trace for the passed in app_name.

    :return: A list of objects.
    """
    async with DbConnection() as db:
        conn_pool = db.get_conn_pool()
        async with conn_pool.acquire() as conn:
            stmt = await conn.prepare(constants.SQL_SELECT_LATEST_RUN)
            async with conn.transaction():
                async for record in stmt.cursor(app_name,
                                                prefetch=_PREFETCH_SIZE):
                    uuid = record['uuid']
        return await _get_trace(uuid, db)


async def get_trace(uuid):
    """Returns all the tracing rows for the passed in uuid.

    :param str uuid: The identifier for the trace run.

    :returns: A list of strings representing a csv view of the tracing run.
    :rtype: list[str]
    """
    async with DbConnection() as db:
        return await _get_trace(uuid, db)


async def _create_tracer(db, identifier, app_name, *column_names):
    """Creates a new run.

    :param str uuid: The uuid of the run expressed as string.
    :param str app_name: The application that is been traced.
    """
    conn_pool = db.get_conn_pool()
    async with conn_pool.acquire() as conn:
        await conn.execute(
            constants.SQL_INSERT_RUN,
            identifier,
            app_name,
            list(column_names))


async def _insert_row(db, uuid, *row_data):
    """Inserts a tracing row to the database.

    :param db: The database object to use.
    :param str uuid: The identifier for the trace run.
    :param row_data: Represent the values for each data point.
    """
    conn_pool = db.get_conn_pool()
    async with conn_pool.acquire() as conn:
        await conn.execute(constants.SQL_INSERT_ROW, uuid, list(row_data))


async def _get_trace(uuid, db):
    """Returns all the tracing rows for the passed in uuid.

    :param str uuid: The identifier for the trace run.

    :returns: A list of strings representing a csv view of the tracing run.
    :rtype: list[str]
    """
    conn_pool = db.get_conn_pool()
    assert uuid
    assert conn_pool
    col_names = []
    async with conn_pool.acquire() as conn:
        stmt = await conn.prepare(constants.SQL_SELECT_COL_NAMES)
        async with conn.transaction():
            async for record in stmt.cursor(uuid, prefetch=_PREFETCH_SIZE):
                col_names.append(record['col_name'])
        clauses = [
            f"row_data[{index + 1}] as {col_name}"
            for index, col_name in enumerate(col_names)
        ]
        sql = "select to_char(date_time, 'YYYY-MM-DD HH24:MI:SS') as timestamp , " + \
              ','.join(clauses) + \
              " from tracing_row where uuid=$1 order by 1"
        stmt = await conn.prepare(sql)

        lines = ['time,' + ','.join(col_names)]
        async with conn.transaction():
            async for record in stmt.cursor(uuid, prefetch=_PREFETCH_SIZE):
                values = [str(v) for v in list(dict(record).values())]
                lines.append(','.join(values))
        return '\n'.join(lines)


async def get_all_tracers():
    tracers = []
    async with DbConnection() as db:
        tracer_names = await _get_tracer_names(db)
        for tracer_name in tracer_names:
            tracers.append(
                {
                    "tracer_name": tracer_name,
                    'runs': await _get_tracer_runs(db, tracer_name)
                }
            )

        return tracers


async def _get_tracer_names(db):
    tracer_names = []
    conn_pool = db.get_conn_pool()
    async with conn_pool.acquire() as conn:
        stmt = await conn.prepare(constants.SQL_SELECT_ALL_TRACER_RUNS)
        async with conn.transaction():
            async for record in stmt.cursor(prefetch=_PREFETCH_SIZE):
                tracer_names.append(record['app_name'])
    return tracer_names


async def _get_tracer_runs(db, app_name):
    tracer_runs = []
    conn_pool = db.get_conn_pool()
    async with conn_pool.acquire() as conn:
        stmt = await conn.prepare(constants.SQL_SELECT_RUNS)
        async with conn.transaction():
            async for record in stmt.cursor(app_name, prefetch=_PREFETCH_SIZE):
                creation_date = record['creation_time']
                tracer_runs.append(
                    {
                        'creation_time': _format_datetime(creation_date),
                        'uuid': record['uuid'],
                    }

                )
    return tracer_runs


def _get_duration(start_time, end_time):
    """Returns the duration from t1 to t2.

    :param datetime.datetime start_time: Start time.
    :param datetime.datetime end_time: End time.

    :return: The duration as a string.
    :rtype: str
    """
    diff = int(math.fabs(int((start_time - end_time).total_seconds())))

    days = diff // constants.DAY_IN_SECONDS
    diff -= days * constants.DAY_IN_SECONDS

    hours = diff // constants.HOUR_IN_SECONDS
    diff -= hours * constants.HOUR_IN_SECONDS

    minutes = diff // constants.MINUTE_IN_SECONDS
    diff -= minutes * constants.MINUTE_IN_SECONDS

    seconds = diff

    tokens = []
    if days:
        tokens.append(f"{days} days")
    if hours:
        tokens.append(f"{hours} hours")
    if minutes:
        tokens.append(f"{minutes} min")
    if seconds:
        tokens.append(f"{seconds} secs")

    return ', '.join(tokens)


def _format_datetime(date_to_format):
    """Formats the passed in date time.

    :param: datetime.datetime date_to_format: The date to format.

    :return: The formatted date.
    :rtype: str
    """
    current_year = datetime.datetime.now().year
    if date_to_format.year == current_year:
        return date_to_format.strftime("%a, %b %d, %H:%M")
    else:
        return date_to_format.strftime("%a, %b %d %Y, %H:%M")



"""Exposes the mnemic UI server."""

import asyncio
import datetime
import functools
import glob
import logging
import os
import uuid

import aiohttp
import aiohttp.web as web
import io
import jinja2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import dolon.utils as utils

logger = logging.getLogger("mnemic")

_PORT = int(os.environ["FRONT_END_PORT"])

_JINJA_ENV = jinja2.Environment(
    loader=jinja2.PackageLoader(
        'templates',
        'templates'),
    autoescape=jinja2.select_autoescape(['html', 'xml'])
)

_CURR_DIR = os.path.dirname(os.path.realpath(__file__))
_IMAGES_DIR = os.path.join(_CURR_DIR, 'static', 'images')
_PATH_TO_STATIC = os.path.join(_CURR_DIR, 'static')
_IMG_URL = ('<img src="/static/images/{image_name}" alt="image n/a" '
            'width="{width}px" height="{height}px">').format
_CONN_STR = f'postgresql://postgres:postgres123@localhost:5432/mnemic'
_IMAGE_CLEANUP_FREQUENCY_IN_SECS = 30
_IMAGE_MAX_LIFE_SPAN_IN_SECS = 10


async def clear_images():
    """Removes unused images."""
    logger.info("Enters image cleaning loop.")
    while 1:
        await asyncio.sleep(_IMAGE_CLEANUP_FREQUENCY_IN_SECS)
        logger.info("Deleting images..")
        files = glob.glob(f'{_IMAGES_DIR}/*.png')
        for f in files:
            time_now = datetime.datetime.now()
            file_time = datetime.datetime.fromtimestamp(os.path.getmtime(f))
            life_span_in_seconds = (time_now - file_time).total_seconds()
            if life_span_in_seconds > _IMAGE_MAX_LIFE_SPAN_IN_SECS:
                try:
                    os.remove(f)
                except Exception as ex:
                    logger.exception(ex)


def web_handler(handler_func):
    """Wraps a handler function adding standard processing."""

    @functools.wraps(handler_func)
    async def _inner(*args, **kwargs):
        """The decorator function."""
        try:
            return await handler_func(*args, **kwargs)
        except Exception as ex:
            logger.exception(ex)
            raise aiohttp.web.HTTPInternalServerError()

    return _inner


def make_image_file():
    """Creates a filename that will be used for an image.

    :return: A filename that will be used for an image.
    :rtype: str.
    """
    image_prefix = str(uuid.uuid4())
    filename = f'{image_prefix}_figure.png'
    return filename


def make_correlation_heat_map(
        data,
        title='Correlation Heat Map',
        linewidths=0,
        figsize=(9, 6),
        annot=False):
    """Creates a correlation heat map.

    :param pandas.DataFrame data: The data to create correlations for.
    :param str title: The title for the correlation image.
    :param int linewidths: The line width for the heat map.
    :param tuple figsize: The figsize as a 2 dimension tuple (in inches).
    :param bool annot: If true the correlation value will be displayed.

    return: The filename of the image and the image itself.
    rtype: tuple.
    """
    cmap = sns.diverging_palette(14, 120, as_cmap=True)
    data = data.dropna()
    corr = data.corr()
    for column_name in corr.columns:
        corr[column_name] = corr[column_name].abs()
    _, ax = plt.subplots(figsize=figsize)
    if title:
        ax.set_title(title)

    heat_map = sns.heatmap(corr, annot=annot, fmt="2.2f",
                           linewidths=linewidths, ax=ax, cmap=cmap)
    figure = heat_map.get_figure()
    figure.set_size_inches(10, 12)
    filename = make_image_file()
    figure.savefig(os.path.join(_IMAGES_DIR, filename), dpi=400)
    return filename, figure


class Handler:
    """Implements all the web handlers used from the service."""

    @web_handler
    async def trace_run_info_handler(self, request):
        """Returns the basic information for the passed in run.

        Expects the uuid of the run which is used to retrieve and return
        a json document consisting of a dict similar to the following:
            {
                'app_name': run-name,
                'counter': number-of-points,
                'started': start-time
                'duration': duration-time
            }

        :param request: The web request.
        """
        uuid_for_run = request.rel_url.query['uuid']
        data = await utils.get_trace_run_info(uuid_for_run)
        return web.json_response(data)

    @web_handler
    async def tracers_handler(self, request):
        """Returns all the available tracer runs sorted by creation time.

        :returns: A json document containing the tracing runs in the
        following format:
            [
               {
                  "tracer_name":"testing_app",
                  "runs":[
                     {
                        "creation_time":"2021-05-29 21:00:20",
                        "uuid":"eefba555-b816-427f-8138-02013067bad8"
                     }
                  ]
               }
            ]

        :param request: The web request.
        """
        data = await utils.get_all_tracers()
        return web.json_response(data)

    @web_handler
    async def tracer_run_handler(self, request):
        """Returns all the rows for the passed in run.

        Expects the uuid of the tracer run to be passed as a query parameter.

        :param request: The web request.
        """
        uuid_for_run = request.rel_url.query['uuid']
        data = await utils.get_trace(uuid_for_run)
        df = pd.read_csv(io.StringIO(data))
        if len(df) == 0:
            # There are no rows for the requested run.
            return web.json_response(
                text="<h1>This tracer does not have any rows to show</h1>",
                content_type='text/html'
            )
        margin_factor = 0.1
        images = []
        image_prefix = str(uuid.uuid4())
        fig_index = 0
        for column_name in df.columns:
            if column_name == 'time':
                continue
            min_value = min(df[column_name])
            max_value = max(df[column_name])
            title = column_name.replace("_", " ").replace("-", " ").title()
            the_plot = df.plot.line(x="time", y=column_name, rot=0,
                                    grid=True,
                                    figsize=(12, 3), title=title,
                                    legend=False)
            if max_value > min_value:
                height = max_value - min_value
                plt.ylim([min_value - height * margin_factor,
                          max_value + height * margin_factor])
            elif max_value < min_value:
                max_value, min_value = min_value, max_value
                height = max_value - min_value
                plt.ylim([min_value - height * margin_factor,
                          max_value + height * margin_factor])
            else:
                height = 0.2
                plt.ylim([min_value - height * margin_factor,
                          max_value + height * margin_factor])
            fig_index += 1
            filename = f'{image_prefix}_figure_{fig_index}.png'
            the_plot.set_facecolor('gainsboro')
            total_number_of_points = len(df)
            x_ticks = np.arange(10, total_number_of_points,
                                total_number_of_points / 12)
            x_ticks = [int(index) for index in x_ticks]
            plt.xticks(x_ticks)
            plt.savefig(os.path.join(_IMAGES_DIR, filename))
            plt.close(the_plot.get_figure())
            images.append(filename)
        correlation_file_name, figure = make_correlation_heat_map(df)
        plt.close(figure)
        html_code = ''
        for image_name in images:
            image_html = _IMG_URL(image_name=image_name, width=1000, height=240)
            html_code += image_html

        html_code += '<div style:"width:1000px;background-color: white">' + _IMG_URL(
            image_name=correlation_file_name, width=600, height=500
        ) + '</div>'

        return web.json_response(
            text=html_code,
            content_type='text/html'
        )

    @web_handler
    async def main_page_handler(self, request):
        """Displays the main page.

        :param request: The web request.
        """
        template = _JINJA_ENV.get_template('index.html')
        txt = template.render()
        return web.Response(
            body=txt.encode(),
            content_type='text/html'
        )


def run():
    """Runs the backend service."""
    app = web.Application()
    handler = Handler()
    app.add_routes(
        [
            web.get('/', handler.main_page_handler),
            web.get('/tracers', handler.tracers_handler),
            web.get('/tracer_run', handler.tracer_run_handler),
            web.get('/trace_run_info', handler.trace_run_info_handler)
        ]
    )
    asyncio.ensure_future(clear_images())
    app.router.add_static('/static', _PATH_TO_STATIC)
    web.run_app(app, port=_PORT)


if __name__ == '__main__':
    run()



"""Implements the function profiler."""

import datetime
import functools
import inspect
import uuid

def clear():
    """Removes all the profiling functions."""
    _ProfilingStatCollection.clear()

def profiler(foo):
    """Decorates a callable or coro making it profile-able.

    :param callable foo: The callable (sync or async) to profile.
    """
    _ProfilingStatCollection.register_profiling_function(foo)
    if inspect.iscoroutinefunction(foo):
        @functools.wraps(foo)
        async def _inner(*args, **kwargs):
            with _ProfilingStatCollection(foo):
                return await foo(*args, **kwargs)
    else:
        @functools.wraps(foo)
        def _inner(*args, **kwargs):
            with _ProfilingStatCollection(foo):
                return foo(*args, **kwargs)
    return _inner


def get_profiling_functions(use_async=False):
    """Returns all the available profiling functions.

    :param bool use_async: If true the returned functions will be async.

    :return: A list of sync/ async async functions one for each function to
        be profiled.  Used by trace client to wire the profile-able functions
        to tracing.

    :rtype: list
    """

    def make_func(profiling_callable_name, profile_func, tag):
        """Dynamically creates a profiling function.

        :param str profiling_callable_name:
        :param callable profile_func:
        :param str tag: The "tag" to use for the name of the returned function.

        :return: A profiled callable customized for the passed in parameters.
        :rtype: callable
        """

        if use_async:
            async def _inner():
                """The function that will be returned.

                :return: A profiling callable.
                :rtype: callable
                """
                return profile_func(profiling_callable_name)
        else:
            def _inner():
                """The function that will be returned.

                :return: A profiling callable.
                :rtype: callable
                """
                return profile_func(profiling_callable_name)

        _inner.__qualname__ = f'{profiling_callable_name}_{tag}'
        _inner.__name__ = f'{profiling_callable_name}_{tag}'

        return _inner

    func = []
    for name in _ProfilingStatCollection.get_profiling_callable_names():
        func.append(make_func(
            profiling_callable_name=name, profile_func=_get_hits, tag='hits')
        )
        func.append(make_func(
            profiling_callable_name=name, profile_func=_get_active_instances,
            tag='active_instances')
        )
        func.append(make_func(
            profiling_callable_name=name, profile_func=_get_average_time,
            tag='average_time')
        )
    return func


def _get_hits(callable_name):
    """Returns the number of hits for the passed-in callable name.

    :param str callable_name: The name of the callable.

    :returns: The number of hits for the passed-in callable name.
    :rtype: int
    """
    return _ProfilingStatCollection.get_stats_for_callable(callable_name).hits


def _get_active_instances(callable_name):
    """Returns the number of running_instances for the passed-in callable name.

    :param str callable_name: The name of the callable.

    :returns: The number of running_instances for the passed-in callable name.
    :rtype: int
    """
    return _ProfilingStatCollection.get_stats_for_callable(
        callable_name).running_instances


def _get_average_time(callable_name):
    """Returns the average_time in seconds for the passed-in callable name.

    :param str callable_name: The name of the callable.

    :returns: The average_time in seconds  for the passed-in callable name.
    :rtype: float
    """
    return _ProfilingStatCollection.get_stats_for_callable(
        callable_name).average_time


class _ProfilingStats:
    """Holds the profiling statistics for a callable.

    Instances of this class are kept in the ProfilingStatCollection as
    a class level dictionary making them available for querying at
    any time.

    :ivar int _hits: How many times a function was called.
    :ivar int _running_instances: How many instances of this functions are
        running (applicable to async calls ofc).
    :ivar float _average_time: The average duration for each call (in seconds).
    :ivar float _enter_time: The time the callable started  (in seconds).
    """

    _hits = 0
    _running_instances = 0
    _average_time = 0
    _enter_time = None

    def __init__(self):
        self._hits = 0
        self._running_instances = 0
        self._average_time = 0
        self._enter_time = {}
        self._completed = 0

    def enter(self, identifier):
        """Called when a callable starts."""
        self._hits += 1
        self._running_instances += 1
        self._enter_time[identifier] = datetime.datetime.now()

    def exit(self, identifier):
        """Called when a callable exits."""
        self._running_instances -= 1
        t_now = self._enter_time.pop(identifier)
        duration = (datetime.datetime.now() - t_now).total_seconds()
        x1 = self._average_time * self._completed + duration
        self._average_time = x1 / (self._completed + 1)
        self._completed += 1

    @property
    def hits(self):
        """Returns the number of hits.

        :return: The number of hits.
        :rtype: int
        """
        return self._hits

    @property
    def running_instances(self):
        """Returns the number of hits.

        :return: The number of concurrent instances.
        :rtype: int
        """
        return self._running_instances

    @property
    def average_time(self):
        """Returns the average duration in seconds.

        :return: The average duration in seconds.
        :rtype: float
        """
        return self._average_time

    def __repr__(self):
        return f'ProfilingStats: hits = {self._hits} ' \
               f'running_instances = {self._running_instances}' \
               f'avg. duration = {self._average_time}'


class _ProfilingStatCollection:
    """Profiling data accumulator.

    A context manager that has keeps a global state where it accumulates
    profiling statistics for callables.

    :cvar dict[str, ProfilingStats]: Maps callables to their profiling stats.

    :ivar str _func_name: The callable's name to profile.
    """

    _stats = {}

    @classmethod
    def clear(cls):
        """Removes all the registered profilers."""
        cls._stats = {}

    @classmethod
    def register_profiling_function(cls, func):
        cls._stats[func.__qualname__] = _ProfilingStats()

    def __init__(self, func):
        """Initializer.

        :param callable func: The callable to profile.
        """
        self._func_name = func.__qualname__
        self._uuid = None

    def __enter__(self):
        """Called when the callable starts."""
        self._uuid = uuid.uuid4()
        self._stats[self._func_name].enter(self._uuid)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Called when the callable exits."""
        self._stats[self._func_name].exit(self._uuid)

    @classmethod
    def get_profiling_callable_names(cls):
        """Returns a list with the names of all callables that are profiled.

        :returns: A list with the names of all callables that are profiled.
        :rtype: list[str]
        """
        return list(cls._stats.keys())

    @classmethod
    def get_stats_for_callable(cls, callable_name):
        """Returns the number of hits for the passed-in callable name.
    
        :param str callable_name: The name of the callable.
        :rtype: _ProfilingStats
        """
        return cls._stats.get(callable_name)



import aiohttp
import asyncio
import datetime
import json
import os
import socket
import tracemalloc
import uuid


class TraceClientImpl:

    def __init__(self, app_name, host, port, verbose, *diagnostics):
        self._app_name = app_name
        self._host = host
        self._port = port
        self._diagnostics = list(diagnostics)
        self._socket = None
        self._verbose = verbose

    def _send(self, data):
        assert self._socket
        assert isinstance(data, dict)
        data["uuid"] = self._uuid
        txt = json.dumps(data)
        self._socket.sendto(txt.encode('utf-8'), (self._host, self._port))

    async def run(self, frequency):
        while 1:
            await asyncio.sleep(frequency)
            msg = {
                "msg_type": "row",
                "row_data": [
                    await diagnostic() for diagnostic in self._diagnostics
                ]
            }
            if self._verbose:
                print("Sending:", msg)
            self._send(msg)

    async def __aenter__(self):
        tracemalloc.start()
        self._uuid = str(uuid.uuid4())
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        column_names = []
        for diagnostic in self._diagnostics:
            if hasattr(diagnostic, "__name__"):
                column_names.append(diagnostic.__name__)
            elif hasattr(diagnostic, "__class__"):
                column_names.append(diagnostic.__class__.__name__)
            else:
                raise ValueError

        msg = {
            "msg_type": "create_trace_run",
            "app_name": self._app_name,
            "column_names": column_names
        }
        self._send(msg)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._socket:
            self._socket.close()
            self._socket = None
            tracemalloc.stop()


class RabbitmqStats:
    """Lower level implementation to retrieve rabbitmq stats."""

    _connections = 0
    _channels = 0
    _queues = 0
    _bindings = 0
    _last_updated = None
    _lock = asyncio.Lock()
    _MIN_UPDATE_INTERVAL = 5
    _DEFAULT_RMQ_SERVER = "http://guest:guest@10.0.2.2:15672"

    @classmethod
    def _get_rmq_server(cls):
        """Returns the RMQ server to use.

        :returns: The RMQ server to use.
        :rtype: str
        """
        return os.environ.get("RMQ_SERVER", cls._DEFAULT_RMQ_SERVER)

    @classmethod
    def _get_min_update_interval(cls):
        return os.environ.get("MIN_RMQ_UPDATE_INTERVAL",
                              cls._MIN_UPDATE_INTERVAL)

    @classmethod
    async def _refresh(cls):
        async with cls._lock:
            now = datetime.datetime.now()
            try:
                if cls._last_updated is None or (
                        now - cls._last_updated).total_seconds() >= cls._MIN_UPDATE_INTERVAL:
                    async with aiohttp.ClientSession() as session:
                        server = cls._get_rmq_server()
                        async with session.get(f'{server}/api/queues') as resp:
                            if 200 <= resp.status < 299:
                                cls._queues = len(await resp.json())

                        async with session.get(f'{server}/api/connections') as resp:
                            if 200 <= resp.status < 299:
                                cls._connections = len(await resp.json())

                        async with session.get(f'{server}/api/channels') as resp:
                            if 200 <= resp.status < 299:
                                cls._channels = len(await resp.json())

                        async with session.get(
                                f'{server}/api/bindings') as resp:
                            if 200 <= resp.status < 299:
                                cls._bindings = len(await resp.json())
                    cls._last_updated = now
            except Exception as ex:
                pass

    @classmethod
    async def rabbitmq_connections(cls):
        """Returns the number of active rabbitmq connections.

        :returns: The number of active rabbitmq connections.

        :rtype: int.
        """
        await cls._refresh()
        return cls._connections

    @classmethod
    async def rabbitmq_channels(cls):
        """Returns the number of active rabbitmq channels.

        :returns: The number of active rabbitmq channels.

        :rtype: int.
        """
        await cls._refresh()
        return cls._channels

    @classmethod
    async def rabbitmq_queues(cls):
        """Returns the number of rabbitmq queues.

        :returns: The number of rabbitmq queues.

        :rtype: int.
        """
        await cls._refresh()
        return cls._queues

    @classmethod
    async def rabbitmq_bindings(cls):
        """Returns the number of rabbitmq bindings.

        :returns: The number of rabbitmq bindings.

        :rtype: int.
        """
        await cls._refresh()
        return cls._bindings



"""Wraps db connection pool within a context manager."""

import os

import dolon.exceptions as exceptions
import dolon.impl.db_conn_impl as db_conn_impl


class DbConnection:
    """Wraps db connection pool within a context manager.

    :ivar DbConnectionImpl _impl: The implementation object.
    :ivar int _min_size: The min number of pooled connections.
    :ivar int _max_size: The max number of pooled connections.
    :ivar str _conn_str: The connection string to the db.
    """

    _impl = None

    _min_size = None
    _max_size = None
    _conn_str = None

    @classmethod
    def _get_conn_str(cls):
        """Gets the connection string using the environment variables.

        Example of connection string:
        'postgresql://postgres:postgres123@localhost:5432/mnemic'

        :returns: The connection string using the environment variables.
        :rtype: str
        """
        if cls._conn_str:
            return cls._conn_str
        else:
            try:
                return os.environ["POSTGRES_CONN_STR"]
            except KeyError as ex:
                raise exceptions.InvalidEnvironmentVariable(str(ex)) from ex

    def __init__(self, min_size=1, max_size=10):
        """Initializer.

        :param int min_size: The min number of pooled connections.
        :param int max_size: The max number of pooled connections.
        """
        self._min_size = min_size
        self._max_size = max_size

    @classmethod
    def set_conn_str(cls, conn_str):
        """Sets the connection string explicitly.

        The explicitly set connection string will take presidency over the
        corresponding environment value.

        :param str conn_str: The connection string to set.
        """
        cls._conn_str = conn_str

    async def __aenter__(self):
        """Enters the context."""
        self._impl = db_conn_impl.DbConnectionImpl(
            min_size=self._min_size,
            max_size=self._max_size,
            conn_str=self._get_conn_str()
        )
        return await self._impl.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exits from the context."""
        if self._impl:
            await self._impl.__aexit__(exc_type, exc_val, exc_tb)
            self._impl = None



"""Exposes the postgress stats."""

_SQL_LIVE_MSGS = """
    select 
        sum(n_live_tup) as counter 
    from 
        pg_stat_user_tables;
"""

_SQL_DEAD_MSGS = """
    select  
        sum(n_dead_tup) as counter 
    from 
        pg_stat_user_tables;
"""

_SQL_IDLE_TRANS = """
SELECT count(*) as counter FROM pg_stat_activity  WHERE state = 'idle';
"""

_SQL_NUMBER_OF_CONNECTIONS = """
SELECT sum(numbackends) as counter FROM pg_stat_database;
"""


class _DbStat:
    """Encapsulates the details for a psql stat."""

    def __init__(self, sql):
        """Initializer.

        :param str sql: The SQL statement to retrieve the stats.
        """
        self._sql = sql

    async def get_value(self, db):
        """Returns the statistic from the passed in database.

        :param DbConnectionImpl db: The db connection object.
        """
        async for value in db.execute_query(self._sql):
            return float(value["counter"])


live_msgs_in_db = _DbStat(_SQL_LIVE_MSGS)
dead_msgs_in_db = _DbStat(_SQL_DEAD_MSGS)
idle_in_db = _DbStat(_SQL_IDLE_TRANS)
conn_count_in_db = _DbStat(_SQL_NUMBER_OF_CONNECTIONS)
